"""Classes for VeSync Switch Devices.

This module provides classes for VeSync Switch Devices:

    1. VeSyncSwitch: Abstract Base class for VeSync Switch Devices.
    2. VeSyncWallSwitch: Class for VeSync Wall Switch Devices ESWL01 and ESWL03.
    3. VeSyncDimmerSwitch: Class for VeSync Dimmer Switch Devices ESWD16.


Attributes:
    feature_dict (dict): Dictionary of switch models and their supported features.
        Defines the class to use for each switch model and the list of features
    switch_modules (dict): Dictionary of switch models as keys and their associated
        classes as string values.

Note:
    The switch device is built from the `feature_dict` dictionary and used by the
    `vesync.object_factory` during initial call to pyvesync.vesync.update() and
    determines the class to instantiate for each switch model. These classes should
    not be instantiated manually.
"""

import logging
import json
from abc import ABCMeta, abstractmethod
from typing import Dict, Union, Optional

from pyvesync.helpers import Helpers as helpers
from pyvesync.vesyncbasedevice import VeSyncBaseDevice

logger = logging.getLogger(__name__)

# --8<-- [start:feature_dict]

feature_dict: Dict[str, Dict[str, Union[list, str]]] = {
    'ESWL01': {
        'module': 'VeSyncWallSwitch',
        'features': []
    },
    'ESWD16': {
        'module': 'VeSyncDimmerSwitch',
        'features': ['dimmable']
    },
    'ESWL03': {
        'module': 'VeSyncWallSwitch',
        'features': []
    }
}

# --8<-- [end:feature_dict]

switch_modules: dict = {k: v['module']
                        for k, v in feature_dict.items()}

__all__: list = list(switch_modules.values()) + ['switch_modules']


class VeSyncSwitch(VeSyncBaseDevice):
    """Etekcity Switch Base Class.

    Abstract Base Class for Etekcity Switch Devices, inherting from
    pyvesync.vesyncbasedevice.VeSyncBaseDevice. Should not be instantiated directly,
    subclassed by VeSyncWallSwitch and VeSyncDimmerSwitch.

    Attributes:
        features (list): List of features supported by the switch device.
        details (dict): Dictionary of switch device details.
    """

    __metaclasss__ = ABCMeta

    def __init__(self, details, manager):
        """Initialize Switch Base Class."""
        super().__init__(details, manager)
        self.features = feature_dict.get(self.device_type, {}).get('features')
        if self.features is None:
            logger.error('% device configuration not set', self.device_name)
            raise KeyError(f'Device configuration not set {self.device_name}')
        self.details = {}

    def is_dimmable(self) -> bool:
        """Return True if switch is dimmable."""
        return bool('dimmable' in self.features)

    @abstractmethod
    def get_details(self) -> None:
        """Get Device Details."""

    @abstractmethod
    def turn_on(self) -> bool:
        """Turn Switch On."""

    @abstractmethod
    def turn_off(self) -> bool:
        """Turn switch off."""

    @abstractmethod
    def get_config(self) -> None:
        """Get configuration and firmware deatils."""

    @property
    def active_time(self) -> int:
        """Get active time of switch."""
        return self.details.get('active_time', 0)

    def update(self) -> None:
        """Update device details."""
        self.get_details()


class VeSyncWallSwitch(VeSyncSwitch):
    """Etekcity standard wall switch class."""

    def __init__(self, details, manager):
        """Initialize standard etekcity wall switch class."""
        super().__init__(details, manager)

    def get_details(self) -> None:
        """Get switch device details."""
        body = helpers.req_body(self.manager, 'devicedetail')
        body['uuid'] = self.uuid
        head = helpers.req_headers(self.manager)

        r, _ = helpers.call_api(
            '/inwallswitch/v1/device/devicedetail', 'post',
            headers=head, json_object=body
        )

        if r is not None and helpers.code_check(r):
            self.device_status = r.get('deviceStatus', self.device_status)
            self.details['active_time'] = r.get('activeTime', 0)
            self.connection_status = r.get(
                'connectionStatus', self.connection_status)
        else:
            logger.debug('Error getting %s details', self.device_name)

    def get_config(self) -> None:
        """Get switch device configuration info."""
        body = helpers.req_body(self.manager, 'devicedetail')
        body['method'] = 'configurations'
        body['uuid'] = self.uuid

        r, _ = helpers.call_api(
            '/inwallswitch/v1/device/configurations',
            'post',
            headers=helpers.req_headers(self.manager),
            json_object=body,
        )

        if helpers.code_check(r):
            self.config = helpers.build_config_dict(r)
        else:
            logger.warning('Unable to get %s config info',
                           self.device_name)

    def turn_off(self) -> bool:
        """Turn off switch device."""
        body = helpers.req_body(self.manager, 'devicestatus')
        body['status'] = 'off'
        body['uuid'] = self.uuid
        head = helpers.req_headers(self.manager)

        r, _ = helpers.call_api(
            '/inwallswitch/v1/device/devicestatus', 'put',
            headers=head, json_object=body
        )

        if r is not None and helpers.code_check(r):
            self.device_status = 'off'
            return True
        logger.warning('Error turning %s off', self.device_name)
        return False

    def turn_on(self) -> bool:
        """Turn on switch device."""
        body = helpers.req_body(self.manager, 'devicestatus')
        body['status'] = 'on'
        body['uuid'] = self.uuid
        head = helpers.req_headers(self.manager)

        r, _ = helpers.call_api(
            '/inwallswitch/v1/device/devicestatus', 'put',
            headers=head, json_object=body
        )

        if r is not None and helpers.code_check(r):
            self.device_status = 'on'
            return True
        logger.warning('Error turning %s on', self.device_name)
        return False


class VeSyncDimmerSwitch(VeSyncSwitch):
    """Vesync Dimmer Switch Class with RGB Faceplate."""

    def __init__(self, details, manager):
        """Initilize dimmer switch class."""
        super().__init__(details, manager)
        self._brightness = 0
        self._rgb_value = {'red': 0, 'blue': 0, 'green': 0}
        self._rgb_status = 'unknown'
        self._indicator_light = 'unknown'

    def get_details(self) -> None:
        """Get dimmer switch details."""
        body = helpers.req_body(self.manager, 'devicedetail')
        body['uuid'] = self.uuid
        head = helpers.req_headers(self.manager)

        r, _ = helpers.call_api(
            '/dimmer/v1/device/devicedetail', 'post',
            headers=head, json_object=body
        )

        if r is not None and helpers.code_check(r):
            self.device_status = r.get('deviceStatus', self.device_status)
            self.details['active_time'] = r.get('activeTime', 0)
            self.connection_status = r.get(
                'connectionStatus', self.connection_status)
            self._brightness = r.get('brightness')
            self._rgb_status = r.get('rgbStatus')
            self._rgb_value = r.get('rgbValue')
            self._indicator_light = r.get('indicatorlightStatus')
        else:
            logger.debug('Error getting %s details', self.device_name)

    @property
    def brightness(self) -> float:
        """Return brightness in percent."""
        return self._brightness

    @property
    def indicator_light_status(self) -> str:
        """Faceplate brightness light status."""
        return self._indicator_light

    @property
    def rgb_light_status(self) -> str:
        """RGB Faceplate light status."""
        return self._rgb_status

    @property
    def rgb_light_value(self) -> dict:
        """RGB Light Values."""
        return self._rgb_value

    def switch_toggle(self, status: str) -> bool:
        """Toggle switch status."""
        if status not in ['on', 'off']:
            logger.debug('Invalid status passed to wall switch')
            return False
        body = helpers.req_body(self.manager, 'devicestatus')
        body['status'] = status
        body['uuid'] = self.uuid
        head = helpers.req_headers(self.manager)

        r, _ = helpers.call_api(
            '/dimmer/v1/device/devicestatus', 'put',
            headers=head, json_object=body
        )

        if r is not None and helpers.code_check(r):
            self.device_status = status
            return True

        logger.warning('Error turning %s %s',
                       self.device_name, status)
        return False

    def turn_on(self) -> bool:
        """Turn switch on."""
        return self.switch_toggle('on')

    def turn_off(self) -> bool:
        """Turn switch off."""
        return self.switch_toggle('off')

    def indicator_light_toggle(self, status: str) -> bool:
        """Toggle indicator light."""
        if status not in ['on', 'off']:
            logger.debug('Invalid status for wall switch')
            return False
        body = helpers.req_body(self.manager, 'devicestatus')
        body['status'] = status
        body['uuid'] = self.uuid
        head = helpers.req_headers(self.manager)

        r, _ = helpers.call_api(
            '/dimmer/v1/device/indicatorlightstatus',
            'put', headers=head, json_object=body
        )

        if r is not None and helpers.code_check(r):
            self.device_status = status
            return True

        logger.warning('Error turning %s indicator light %s',
                       self.device_name, status)
        return False

    def indicator_light_on(self) -> bool:
        """Turn Indicator light on."""
        return self.indicator_light_toggle('on')

    def indicator_light_off(self) -> bool:
        """Turn indicator light off."""
        return self.indicator_light_toggle('off')

    def rgb_color_status(self, status: str,
                         red: Optional[int] = None,
                         blue: Optional[int] = None,
                         green: Optional[int] = None) -> bool:
        """Set faceplate RGB color."""
        body = helpers.req_body(self.manager, 'devicestatus')
        body['status'] = status
        body['uuid'] = self.uuid
        head = helpers.req_headers(self.manager)
        if red is not None and blue is not None and green is not None:
            body['rgbValue'] = {'red': red, 'blue': blue, 'green': green}

        r, _ = helpers.call_api(
            '/dimmer/v1/device/devicergbstatus', 'put',
            headers=head, json_object=body
        )

        if r is not None and helpers.code_check(r):
            self._rgb_status = status
            if body.get('rgbValue') is not None:
                self._rgb_value = {'red': red, 'blue': blue, 'green': green}
            return True
        logger.warning('Error turning %s off', self.device_name)
        return False

    def rgb_color_off(self) -> bool:
        """Turn RGB Color Off."""
        return self.rgb_color_status('off')

    def rgb_color_on(self) -> bool:
        """Turn RGB Color Off."""
        return self.rgb_color_status('on')

    def rgb_color_set(self, red: int, green: int, blue: int) -> bool:
        """Set RGB color of faceplate."""
        try:
            red = int(red)
            green = int(green)
            blue = int(blue)
        except ValueError:
            return False
        if isinstance(red, int) and isinstance(
                green, int) and isinstance(blue, int):
            for color in [red, green, blue]:
                if color < 0 or color > 255:
                    logger.warning('Invalid RGB value')
                    return False

            return bool(self.rgb_color_status('on', red, green, blue))
        return False

    def set_brightness(self, brightness: int) -> bool:
        """Set brightness of dimmer - 1 - 100."""
        if isinstance(brightness, int) and (
                brightness > 0 or brightness <= 100):

            body = helpers.req_body(self.manager, 'devicestatus')
            body['brightness'] = brightness
            body['uuid'] = self.uuid
            head = helpers.req_headers(self.manager)

            r, _ = helpers.call_api(
                '/dimmer/v1/device/updatebrightness', 'put',
                headers=head, json_object=body
            )

            if r is not None and helpers.code_check(r):
                self._brightness = brightness
                return True
            logger.warning('Error setting %s brightness', self.device_name)
        else:
            logger.warning('Invalid brightness')
        return False

    def displayJSON(self) -> str:
        """JSON API for dimmer switch."""
        sup_val = json.loads(super().displayJSON())
        if self.is_dimmable is True:  # pylint: disable=using-constant-test
            sup_val.update(
                {
                    'Indicator Light': str(self.active_time),
                    'Brightness': str(self._brightness),
                    'RGB Light': str(self._rgb_status),
                }
            )
        return json.dumps(sup_val, indent=4)

    def get_config(self) -> None:
        """Get dimmable switch device configuration info."""
        body = helpers.req_body(self.manager, 'devicedetail')
        body['method'] = 'configurations'
        body['uuid'] = self.uuid

        r, _ = helpers.call_api(
            '/dimmer/v1/device/configurations',
            'post',
            headers=helpers.req_headers(self.manager),
            json_object=body,
        )

        if helpers.code_check(r):
            self.config = helpers.build_config_dict(r)
        else:
            logger.warning('Unable to get %s config info', self.device_name)



"""
Air Purifier and Humidifier Device API Responses

FANS variable is a list of device types

DETAILS_RESPONSES variable is a dictionary of responses from the API
for get_details() methods.  The keys are the device types and the
values are the responses.  The responses are tuples of (response, status)

METHOD_RESPONSES variable is a defaultdict of responses from the API. This is
the FunctionResponse variable from the utils module in the tests dir.
The default response is a tuple with the value ({"code": 0, "msg": "success"}, 200).

The values of METHOD_RESPONSES can be a function that takes a single argument or
a static value. The value is checked if callable at runtime and if so, it is called
with the provided argument. If not callable, the value is returned as is.

METHOD_RESPONSES = {
    'DEV_TYPE': defaultdict(
        lambda: ({"code": 0, "msg": "success"}, 200))
    )
}

# For a function to handle the response
def status_response(request_body=None):
    # do work with request_body
    return request_body, 200

METHOD_RESPONSES['DEV_TYPE']['set_status'] = status_response

# To change the default value for a device type

METHOD_RESPONSES['DEVTYPE'].default_factory = lambda: ({"code": 0, "msg": "success"}, 200)
"""

from copy import deepcopy
from pyvesync import vesyncfan, helpers
from utils import Defaults, FunctionResponses

HUMID_MODELS = []
for model_type, model_dict in vesyncfan.humid_features.items():
    HUMID_MODELS.append(model_dict["models"][0])

AIR_MODELS = []
for model_type, model_dict in vesyncfan.air_features.items():
    AIR_MODELS.append(model_dict["models"][0])

FANS = HUMID_MODELS + AIR_MODELS
FANS_NUM = len(FANS)
# FANS = ['Core200S', 'Core300S', 'Core400S', 'Core600S', 'LV-PUR131S', 'LV600S',
#         'Classic300S', 'Classic200S', 'Dual200S', 'LV600S']


def INNER_RESULT(inner: dict) -> dict:
    return {
        "traceId": Defaults.trace_id,
        "code": 0,
        "msg": "request success",
        "module": None,
        "stacktrace": None,
        "result": {"traceId": Defaults.trace_id, "code": 0, "result": inner},
    }


class FanDefaults:
    fan_level = 1
    filter_life = 80
    humidity = 50
    mist_level = 3
    warm_mist_level = 2
    air_quality = 3
    air_quality_value = 4
    filter_open = 0
    aq_percent = 75
    pm1 = 10
    pm10 = 5
    rotate_angle = 45


class FanDetails:
    details_air = ({
        "code": 0,
        "msg": None,
        "deviceStatus": "on",
        "connectionStatus": "online",
        "activeTime": Defaults.active_time,
        "deviceImg": None,
        "deviceName": "LV-PUR131S-NAME",
        "filterLife": {
            "change": False,
            "useHour": None,
            "percent": FanDefaults.filter_life,
        },
        "airQuality": "excellent",
        "screenStatus": "on",
        "mode": "manual",
        "level": FanDefaults.fan_level,
        "schedule": None,
        "timer": None,
        "scheduleCount": 0,
    }, 200)

    details_lv600s = ({
        "traceId": Defaults.trace_id,
        "code": 0,
        "msg": "请求成功",
        "result": {
            "traceId": Defaults.trace_id,
            "code": 0,
            "result": {
                "enabled": True,
                "humidity": FanDefaults.humidity,
                "mist_virtual_level": 8,
                "mist_level": FanDefaults.mist_level,
                "mode": "manual",
                "water_lacks": False,
                "humidity_high": False,
                "water_tank_lifted": False,
                "display": False,
                "automatic_stop_reach_target": True,
                "night_light_brightness": 0,
                "warm_mist_level": 0,
                "warm_mist_enabled": False,
                "configuration": {
                    "auto_target_humidity": 50,
                    "display": False,
                    "automatic_stop": True,
                },
            },
        },
    }, 200)

    details_classic200s300s = ({
        "traceId": Defaults.trace_id,
        "code": 0,
        "msg": "请求成功",
        "result": {
            "traceId": Defaults.trace_id,
            "code": 0,
            "result": {
                "enabled": True,
                "humidity": FanDefaults.humidity,
                "mist_virtual_level": 8,
                "mist_level": FanDefaults.mist_level,
                "mode": "manual",
                "water_lacks": False,
                "humidity_high": False,
                "water_tank_lifted": False,
                "display": False,
                "automatic_stop_reach_target": True,
                "night_light_brightness": 0,
                "configuration": {
                    "auto_target_humidity": 50,
                    "display": False,
                    "automatic_stop": True,
                },
            },
        },
    }, 200)

    details_oasismist1000S = ({
        "traceId": Defaults.trace_id,
        "code": 0,
        "msg": "request success",
        "result": {
            "traceId": Defaults.trace_id,
            "code": 0,
            "result": {
                "powerSwitch": 0,
                "humidity": FanDefaults.humidity,
                "targetHumidity": 50,
                "virtualLevel": 1,
                "mistLevel": FanDefaults.mist_level,
                "workMode": "manual",
                "waterLacksState": 0,
                "waterTankLifted": 0,
                "autoStopSwitch": 1,
                "autoStopState": 0,
                "screenSwitch": 1,
                "screenState": 0,
                "scheduleCount": 0,
                "timerRemain": 0,
                "errorCode": 0,
            },
        },
    }, 200)

    details_core = ({

        "traceId": Defaults.trace_id,
        "code": 0,
        "msg": "request success",
        "result": {
            "traceId": Defaults.trace_id,
            "code": 0,
            "result": {
                "enabled": True,
                "filter_life": FanDefaults.filter_life,
                "mode": "manual",
                "level": FanDefaults.fan_level,
                "air_quality": FanDefaults.air_quality,
                "air_quality_value": FanDefaults.air_quality_value,
                "display": True,
                "child_lock": True,
                "configuration": {
                    "display": True,
                    "display_forever": True,
                    "auto_preference": {"type": "default", "room_size": 0},
                },
                "extension": {"schedule_count": 0, "timer_remain": 0},
                "device_error_code": 0,
            },
        },
    }, 200)

    details_vital100s = ({
        "traceId": Defaults.trace_id,
        "code": 0,
        "msg": "request success",
        "module": None,
        "stacktrace": None,
        "result": {
            "traceId": Defaults.trace_id,
            "code": 0,
            "result": {
                "powerSwitch": Defaults.bin_toggle,
                "filterLifePercent": FanDefaults.filter_life,
                "workMode": "manual",
                "manualSpeedLevel": FanDefaults.fan_level,
                "fanSpeedLevel": FanDefaults.fan_level,
                "AQLevel": FanDefaults.air_quality,
                "PM25": FanDefaults.air_quality_value,
                "screenState": Defaults.bin_toggle,
                "childLockSwitch": Defaults.bin_toggle,
                "screenSwitch": Defaults.bin_toggle,
                "lightDetectionSwitch": Defaults.bin_toggle,
                "environmentLightState": Defaults.bin_toggle,
                "autoPreference": {"autoPreferenceType": "default", "roomSize": 0},
                "scheduleCount": 0,
                "timerRemain": 0,
                "efficientModeTimeRemain": 0,
                "sleepPreference": {
                    "sleepPreferenceType": "default",
                    "cleaningBeforeBedSwitch": 1,
                    "cleaningBeforeBedSpeedLevel": 3,
                    "cleaningBeforeBedMinutes": 5,
                    "whiteNoiseSleepAidSwitch": 1,
                    "whiteNoiseSleepAidSpeedLevel": 1,
                    "whiteNoiseSleepAidMinutes": 45,
                    "duringSleepSpeedLevel": 5,
                    "duringSleepMinutes": 480,
                    "afterWakeUpPowerSwitch": 1,
                    "afterWakeUpWorkMode": "auto",
                    "afterWakeUpFanSpeedLevel": 1,
                },
                "errorCode": 0,
            },
        },
    }, 200)

    details_superior6000S = ({
        "traceId": Defaults.trace_id,
        "code": 0,
        "msg": "request success",
        "module": None,
        "stacktrace": None,
        "result": {
            "traceId": Defaults.trace_id,
            "code": 0,
            "result": {
                "powerSwitch": 1,
                "humidity": 44,
                "targetHumidity": 50,
                "virtualLevel": 1,
                "mistLevel": 1,
                "workMode": "manual",
                "waterLacksState": 0,
                "waterTankLifted": 0,
                "autoStopSwitch": 1,
                "autoStopState": 0,
                "screenSwitch": 1,
                "screenState": 1,
                "scheduleCount": 0,
                "timerRemain": 0,
                "errorCode": 0,
                "dryingMode": {
                    "dryingLevel": 1,
                    "autoDryingSwitch": 1,
                    "dryingState": 2,
                    "dryingRemain": 7200,
                },
                "autoPreference": 1,
                "childLockSwitch": 0,
                "filterLifePercent": 93,
                "temperature": 662,
            },
        },
    }, 200)

    details_everest = ({
        "traceId": "1691789977402",
        "code": 0,
        "msg": "request success",
        "module": None,
        "stacktrace": None,
        "result": {
            "traceId": Defaults.trace_id,
            "code": 0,
            "result": {
                "fanRotateAngle": FanDefaults.rotate_angle,
                "filterOpenState": FanDefaults.filter_open,
                "powerSwitch": Defaults.bin_toggle,
                "filterLifePercent": FanDefaults.filter_life,
                "workMode": "auto",
                "manualSpeedLevel": FanDefaults.fan_level,
                "fanSpeedLevel": FanDefaults.fan_level,
                "AQLevel": FanDefaults.air_quality,
                "AQPercent": FanDefaults.aq_percent,
                "PM25": FanDefaults.air_quality_value,
                "PM1": FanDefaults.pm1,
                "PM10": FanDefaults.pm10,
                "screenState": Defaults.bin_toggle,
                "childLockSwitch": Defaults.bin_toggle,
                "screenSwitch": Defaults.bin_toggle,
                "lightDetectionSwitch": Defaults.bin_toggle,
                "environmentLightState": Defaults.bin_toggle,
                "autoPreference": {"autoPreferenceType": "default", "roomSize": 0},
                "routine": {"routineType": "normal", "runSeconds": 0},
                "scheduleCount": 0,
                "timerRemain": 0,
                "efficientModeTimeRemain": 0,
                "ecoModeRunTime": 0,
                "errorCode": 0,
            },
        },
    }, 200)

    details_towerfan42 = ({
        "traceId": Defaults.trace_id,
        "code": 0,
        "msg": "request success",
        "module": None,
        "stacktrace": None,
        "result": {
            "traceId": Defaults.trace_id,
            "code": 0,
            "result": {
                "powerSwitch": Defaults.bin_toggle,
                "workMode": "turbo",
                "manualSpeedLevel": 3,
                "fanSpeedLevel": 12,
                "screenState": Defaults.bin_toggle,
                "screenSwitch": Defaults.bin_toggle,
                "oscillationSwitch": Defaults.bin_toggle,
                "oscillationState": Defaults.bin_toggle,
                "muteSwitch": Defaults.bin_toggle,
                "muteState": Defaults.bin_toggle,
                "timerRemain": 0,
                "temperature": 750,
                "sleepPreference": {
                  "sleepPreferenceType": "default",
                  "oscillationSwitch": Defaults.bin_toggle,
                  "initFanSpeedLevel": 0,
                  "fallAsleepRemain": 0,
                  "autoChangeFanLevelSwitch": Defaults.bin_toggle
                },
                "scheduleCount": 0,
                "displayingType": 0,
                "errorCode": 0,
            }                
        },
    }, 200)


DETAILS_RESPONSES = {
    "LV-PUR131S": FanDetails.details_air,
    "Classic300S": FanDetails.details_classic200s300s,
    "Classic200S": FanDetails.details_classic200s300s,
    "Dual200S": FanDetails.details_classic200s300s,
    "LUH-A602S-WUSR": FanDetails.details_lv600s,
    "Core200S": FanDetails.details_core,
    "Core300S": FanDetails.details_core,
    "Core400S": FanDetails.details_core,
    "Core600S": FanDetails.details_core,
    "LUH-O451S-WEU": FanDetails.details_lv600s,
    "LUH-O451S-WUS": FanDetails.details_lv600s,
    "LAP-V201S-AASR": FanDetails.details_vital100s,
    "LAP-V102S-AASR": FanDetails.details_vital100s,
    "LUH-M101S-WUS": FanDetails.details_oasismist1000S,
    "LEH-S601S-WUS": FanDetails.details_superior6000S,
    "LAP-EL551S-AUS": FanDetails.details_everest,
    "LTF-F422S-KEU": FanDetails.details_towerfan42,
}

FunctionResponses.default_factory = lambda: (
    {
        "traceId": Defaults.trace_id,
        "code": 0,
        "msg": "request success",
        "result": {"traceId": Defaults.trace_id, "code": 0},
    },
    200,
)

METHOD_RESPONSES = {k: deepcopy(FunctionResponses) for k in FANS}

# Add responses for methods with different responses than the default

# Timer Responses

for k in AIR_MODELS:
    METHOD_RESPONSES[k]["set_timer"] = (INNER_RESULT({"id": 1}), 200)
    METHOD_RESPONSES[k]["get_timer"] = (
        INNER_RESULT({"id": 1, "remain": 100, "total": 100, "action": "off"}),
        200,
    )
FAN_TIMER = helpers.Timer(100, "off")



import pytest
from unittest.mock import patch
import logging
from pyvesync import VeSync
from pyvesync.vesyncswitch import VeSyncWallSwitch
from pyvesync.helpers import Helpers as helpers
import call_json
import call_json_switches

DEVICE_TYPE = 'ESWL01'

DEV_LIST_DETAIL = call_json.DeviceList.device_list_item(DEVICE_TYPE)

CORRECT_WS_LIST = call_json.DeviceList.device_list_response(DEVICE_TYPE)

CORRECT_WS_DETAILS = call_json_switches.DETAILS_RESPONSES[DEVICE_TYPE]

BAD_LIST = call_json.DETAILS_BADCODE

DEFAULTS = call_json.Defaults


class TestVesyncWallSwitch(object):
    @pytest.fixture()
    def api_mock(self, caplog):
        self.mock_api_call = patch('pyvesync.helpers.Helpers.call_api')
        self.mock_api = self.mock_api_call.start()
        self.mock_api.create_autospect()
        self.mock_api.return_value.ok = True
        self.vesync_obj = VeSync('sam@mail.com', 'pass', debug=True)
        self.vesync_obj.enabled = True
        self.vesync_obj.login = True
        self.vesync_obj.token = DEFAULTS.token
        self.vesync_obj.account_id = DEFAULTS.account_id
        caplog.set_level(logging.DEBUG)
        yield
        self.mock_api_call.stop()

    def test_ws_conf(self, api_mock):
        """Tests that Wall Switch is instantiated properly"""
        self.mock_api.return_value = CORRECT_WS_LIST
        self.vesync_obj.get_devices()
        switch = self.vesync_obj.switches
        assert len(switch) == 1
        wswitch = switch[0]
        assert isinstance(wswitch, VeSyncWallSwitch)
        assert wswitch.device_name == DEFAULTS.name(DEVICE_TYPE)
        assert wswitch.device_type == DEVICE_TYPE
        assert wswitch.cid == DEFAULTS.cid(DEVICE_TYPE)
        assert wswitch.uuid == DEFAULTS.uuid(DEVICE_TYPE)

    def test_ws_details(self, api_mock):
        """Test WS get_details() """
        self.mock_api.return_value = CORRECT_WS_DETAILS
        wswitch = VeSyncWallSwitch(DEV_LIST_DETAIL, self.vesync_obj)
        wswitch.get_details()
        dev_details = wswitch.details
        assert wswitch.device_status == 'on'
        assert type(dev_details) == dict
        assert dev_details['active_time'] == 1
        assert wswitch.connection_status == 'online'

    def test_ws_details_fail(self, caplog, api_mock):
        """Test WS get_details with Code>0"""
        self.mock_api.return_value = BAD_LIST
        vswitch15a = VeSyncWallSwitch(DEV_LIST_DETAIL, self.vesync_obj)
        vswitch15a.get_details()
        assert len(caplog.records) == 1
        assert 'details' in caplog.text

    def test_ws_onoff(self, caplog, api_mock):
        """Test 15A Device On/Off Methods"""
        self.mock_api.return_value = ({'code': 0}, 200)
        wswitch = VeSyncWallSwitch(DEV_LIST_DETAIL, self.vesync_obj)
        head = helpers.req_headers(self.vesync_obj)
        body = helpers.req_body(self.vesync_obj, 'devicestatus')

        body['status'] = 'on'
        body['uuid'] = wswitch.uuid
        on = wswitch.turn_on()
        self.mock_api.assert_called_with(
            '/inwallswitch/v1/device/devicestatus', 'put', headers=head, json_object=body
        )
        assert on
        off = wswitch.turn_off()
        body['status'] = 'off'
        self.mock_api.assert_called_with(
            '/inwallswitch/v1/device/devicestatus', 'put', headers=head, json_object=body
        )
        assert off

    def test_ws_onoff_fail(self, api_mock):
        """Test ws On/Off Fail with Code>0"""
        self.mock_api.return_value = ({'code': 1}, 400)
        vswitch15a = VeSyncWallSwitch(DEV_LIST_DETAIL, self.vesync_obj)
        assert not vswitch15a.turn_on()
        assert not vswitch15a.turn_off()



"""VeSync Kitchen Devices."""
import json
import logging
import time
from functools import wraps
from typing import Optional, Union, Set
from dataclasses import dataclass
from pyvesync.vesyncbasedevice import VeSyncBaseDevice
from pyvesync.helpers import Helpers as helpers

logger = logging.getLogger(__name__)


kitchen_features: dict = {
    'Cosori3758L': {
        'module': 'VeSyncAirFryer158',
        'models': ['CS137-AF/CS158-AF', 'CS158-AF', 'CS137-AF', 'CS358-AF'],
        'features': [],
    }
}


def model_dict() -> dict:
    """Build purifier and humidifier model dictionary."""
    model_modules = {}
    for dev_dict in kitchen_features.values():
        for model in dev_dict['models']:
            model_modules[model] = dev_dict['module']
    return model_modules


def model_features(dev_type: str) -> dict:
    """Get features from device type."""
    for dev_dict in kitchen_features.values():
        if dev_type in dev_dict['models']:
            return dev_dict
    raise ValueError('Device not configured')


kitchen_classes: Set[str] = {v['module'] for k, v in kitchen_features.items()}


kitchen_modules: dict = model_dict()


__all__ = list(kitchen_classes)


# Status refresh interval in seconds
# API calls outside of interval are automatically refreshed
# Set VeSyncAirFryer158.refresh_interval to 0 to refresh every call
# Set to None or -1 to disable auto-refresh
REFRESH_INTERVAL = 60

RECIPE_ID = 1
RECIPE_TYPE = 3
CUSTOM_RECIPE = 'Manual Cook'
COOK_MODE = 'custom'


def check_status(func):
    """Check interval between updates."""
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        seconds_elapsed = int(time.time()) - self.last_update
        logger.debug("Seconds elapsed between updates: %s", seconds_elapsed)
        refresh = False
        if self.refresh_interval is None:
            refresh = bool(seconds_elapsed > REFRESH_INTERVAL)
        elif self.refresh_interval == 0:
            refresh = True
        elif self.refresh_interval > 0:
            refresh = bool(seconds_elapsed > self.refresh_interval)
        if refresh is True:
            logger.debug("Updating status, %s seconds elapsed", seconds_elapsed)
            self.update()
        return func(self, *args, **kwargs)
    return wrapper


@dataclass(init=False, eq=False, repr=False)
class FryerStatus:
    """Dataclass for air fryer status."""

    ready_start: bool = False
    preheat: bool = False
    cook_status: Optional[str] = None
    current_temp: Optional[int] = None
    cook_set_temp: Optional[int] = None
    cook_set_time: Optional[int] = None
    cook_last_time: Optional[int] = None
    last_timestamp: Optional[int] = None
    preheat_set_time: Optional[int] = None
    preheat_last_time: Optional[int] = None
    _temp_unit: Optional[str] = None

    @property
    def is_resumable(self) -> bool:
        """Return if cook is resumable."""
        if self.cook_status in ['cookStop', 'preheatStop']:
            if self.cook_set_time is not None:
                return self.cook_set_time > 0
            if self.preheat_set_time is not None:
                return self.preheat_set_time > 0
        return False

    @property
    def temp_unit(self) -> Optional[str]:
        """Return temperature unit."""
        return self._temp_unit

    @temp_unit.setter
    def temp_unit(self, temp_unit: str):
        """Set temperature unit."""
        if temp_unit.lower() in ['f', 'fahrenheit']:
            self._temp_unit = 'fahrenheit'
        elif temp_unit.lower() in ['c', 'celsius']:
            self._temp_unit = 'celsius'
        else:
            raise ValueError(f'Invalid temperature unit - {temp_unit}')

    @property
    def preheat_time_remaining(self) -> int:
        """Return preheat time remaining."""
        if self.preheat is False or self.cook_status == 'preheatEnd':
            return 0
        if self.cook_status in ['pullOut', 'preheatStop']:
            if self.preheat_last_time is None:
                return 0
            return int(self.preheat_last_time // 60)
        if self.preheat_last_time is not None and self.last_timestamp is not None:
            return int(max((self.preheat_last_time -
                            (int(time.time()) - self.last_timestamp)) // 60, 0))
        return 0

    @property
    def cook_time_remaining(self) -> int:
        """Returns the amount of time remaining if cooking."""
        if self.preheat is True or self.cook_status == 'cookEnd':
            return 0
        if self.cook_status in ['pullOut', 'cookStop']:
            if self.cook_last_time is None:
                return 0
            return int(max(self.cook_last_time // 60, 0))
        if self.cook_last_time is not None and self.last_timestamp is not None:
            return int(max((self.cook_last_time -
                            (int(time.time()) - self.last_timestamp)) // 60, 0))
        return 0

    @property
    def remaining_time(self):
        """Return minutes remaining if cooking/heating."""
        if self.preheat is True:
            return self.preheat_time_remaining
        return self.cook_time_remaining

    @property
    def is_running(self) -> bool:
        """Return if cooking or heating."""
        return bool(self.cook_status in ['cooking', 'heating']) and \
            bool(self.remaining_time > 0)

    @property
    def is_cooking(self) -> bool:
        """Return if cooking."""
        return self.cook_status == 'cooking' and self.remaining_time > 0

    @property
    def is_heating(self) -> bool:
        """Return if heating."""
        return self.cook_status == 'heating' and self.remaining_time > 0

    def status_request(self, json_cmd: dict) -> None:  # pylint: disable=R1260
        """Set status from jsonCmd of API call."""
        self.last_timestamp = None
        if not isinstance(json_cmd, dict):
            return
        self.preheat = False
        preheat = json_cmd.get('preheat')
        cook = json_cmd.get('cookMode')
        if isinstance(preheat, dict):
            self.preheat = True
            if preheat.get('preheatStatus') == 'stop':
                self.cook_status = 'preheatStop'
            elif preheat.get('preheatStatus') == 'heating':
                self.cook_status = 'heating'
                self.last_timestamp = int(time.time())
                self.preheat_set_time = preheat.get('preheatSetTime',
                                                    self.preheat_set_time)
                if preheat.get('preheatSetTime') is not None:
                    self.preheat_last_time = preheat.get('preheatSetTime')
                self.cook_set_temp = preheat.get('targetTemp', self.cook_set_temp)
                self.cook_set_time = preheat.get('cookSetTime', self.cook_set_time)
                self.cook_last_time = None
            elif preheat.get('preheatStatus') == 'end':
                self.cook_status = 'preheatEnd'
                self.preheat_last_time = 0
        elif isinstance(cook, dict):
            self.clear_preheat()
            if cook.get('cookStatus') == 'stop':
                self.cook_status = 'cookStop'
            elif cook.get('cookStatus') == 'cooking':
                self.cook_status = 'cooking'
                self.last_timestamp = int(time.time())
                self.cook_set_time = cook.get('cookSetTime', self.cook_set_time)
                self.cook_set_temp = cook.get('cookSetTemp', self.cook_set_temp)
                self.current_temp = cook.get('currentTemp', self.current_temp)
                self.temp_unit = cook.get('tempUnit', self.temp_unit)
            elif cook.get('cookStatus') == 'end':
                self.set_standby()
                self.cook_status = 'cookEnd'

    def clear_preheat(self):
        """Clear preheat status."""
        self.preheat = False
        self.preheat_set_time = None
        self.preheat_last_time = None

    def set_standby(self):
        """Clear cooking status."""
        self.cook_status = 'standby'
        self.clear_preheat()
        self.cook_last_time = None
        self.current_temp = None
        self.cook_set_time = None
        self.cook_set_temp = None
        self.last_timestamp = None

    def status_response(self, return_status: dict) -> None:
        """Set status of Air Fryer Based on API Response."""
        self.last_timestamp = None
        self.preheat = False
        self.cook_status = return_status.get('cookStatus')
        if self.cook_status == 'standby':
            self.set_standby()
            return

        #  If drawer is pulled out, set standby if resp does not contain other details
        if self.cook_status == 'pullOut':
            self.last_timestamp = None
            if 'currentTemp' not in return_status or 'tempUnit' not in return_status:
                self.set_standby()
                self.cook_status = 'pullOut'
                return
        if return_status.get('preheatLastTime') is not None or \
                self.cook_status in ['heating', 'preheatStop', 'preheatEnd']:
            self.preheat = True

        self.cook_set_time = return_status.get('cookSetTime', self.cook_set_time)
        self.cook_last_time = return_status.get('cookLastTime')
        self.current_temp = return_status.get('curentTemp')
        self.cook_set_temp = return_status.get('targetTemp',
                                               return_status.get('cookSetTemp'))
        self.temp_unit = return_status.get('tempUnit', self.temp_unit)  # Always keep set
        self.preheat_set_time = return_status.get('preheatSetTime')
        self.preheat_last_time = return_status.get('preheatLastTime')

        #  Set last_time timestamp if cooking
        if self.cook_status in ['cooking', 'heating']:
            self.last_timestamp = int(time.time())

        if self.cook_status == 'preheatEnd':
            self.preheat_last_time = 0
            self.cook_last_time = None
        if self.cook_status == 'cookEnd':
            self.cook_last_time = 0

        #  If Cooking, clear preheat status
        if self.cook_status in ['cooking', 'cookStop', 'cookEnd']:
            self.clear_preheat()


class VeSyncAirFryer158(VeSyncBaseDevice):
    """Cosori Air Fryer Class."""

    def __init__(self, details, manager):
        """Init the VeSync Air Fryer 158 class."""
        super().__init__(details, manager)
        self.fryer_status = FryerStatus()

        if self.pid is None:
            self.get_pid()
        self.fryer_status.temp_unit = self.get_temp_unit()
        self.ready_start = self.get_remote_cook_mode()
        self.last_update = int(time.time())
        self.refresh_interval = 0

    def get_body(self, method:  Optional[str] = None) -> dict:
        """Return body of api calls."""
        body = {
            **helpers.req_body(self.manager, 'bypass'),
            'cid': self.cid,
            'userCountryCode': self.manager.country_code,
            'debugMode': False
        }
        if method is not None:
            body['method'] = method
        return body

    def get_status_body(self, cmd_dict: dict) -> dict:
        """Return body of api calls."""
        body = self.get_body()
        body.update({
            'uuid': self.uuid,
            'configModule': self.config_module,
            'jsonCmd': cmd_dict,
            'pid': self.pid,
            'accountID': self.manager.account_id
            }
        )
        return body

    def get_temp_unit(self):
        """Get Air Fryer Configuration."""
        body = self.get_body('configurationsV2')

        r, _ = helpers.call_api('/cloud/v2/deviceManaged/configurationsV2', 'post',
                                json_object=body)
        if not isinstance(r, dict) or r.get('code') != 0 \
                or not isinstance(r.get('result'), dict):
            logger.debug('Failed to get config for %s', self.device_name)
            return
        result = r.get('result')
        if result is not None:
            return result.get('airFryerInfo', {}).get('workTempUnit', 'f')

    def get_remote_cook_mode(self):
        """Get the cook mode."""
        body = self.get_body('getRemoteCookMode158')
        r, _ = helpers.call_api('/cloud/v1/deviceManaged/getRemoteCookMode158',
                                'post',
                                json_object=body)
        if not isinstance(r, dict) or r.get('code') != 0 \
                or not isinstance(r.get('result'), dict):
            return False
        return r.get('result', {}).get('readyStart', False)

    @property
    def temp_unit(self) -> Optional[str]:
        """Return temp unit."""
        return self.fryer_status.temp_unit

    @property
    def current_temp(self) -> Optional[int]:
        """Return current temperature."""
        return self.fryer_status.current_temp

    @property
    def cook_set_temp(self) -> Optional[int]:
        """Return set temperature."""
        return self.fryer_status.cook_set_temp

    @property
    def preheat(self) -> bool:
        """Return preheat status."""
        return self.fryer_status.preheat

    @property
    def cook_last_time(self) -> Optional[int]:
        """Return cook last time."""
        return self.fryer_status.cook_last_time

    @property
    def cook_set_time(self) -> Optional[int]:
        """Return cook set time."""
        return self.fryer_status.cook_set_time

    @property
    def preheat_last_time(self) -> Optional[int]:
        """Return preheat last time."""
        if self.preheat is True:
            return self.fryer_status.preheat_last_time
        return None

    @property
    def preheat_set_time(self) -> Optional[int]:
        """Return preheat set time."""
        if self.preheat is True:
            return self.fryer_status.preheat_set_time
        return None

    @property
    def cook_status(self) -> Optional[str]:
        """Return the cook status."""
        return self.fryer_status.cook_status

    @property
    def is_cooking(self) -> bool:
        """Return True if air fryer is heating."""
        return self.fryer_status.is_cooking

    @property
    def is_heating(self) -> bool:
        """Return True if air fryer is preheating."""
        return self.fryer_status.is_heating

    @property
    def is_running(self):
        """Return True if cooking or preheating finished."""
        return self.fryer_status.is_running

    @property
    def remaining_time(self) -> Optional[int]:
        """Return time remaining in minutes or None if not cooking/heating."""
        return self.fryer_status.remaining_time

    def get_details(self):
        """Get Air Fryer Status and Details."""
        cmd = {'getStatus': 'status'}
        req_body = self.get_status_body(cmd)
        url = '/cloud/v1/deviceManaged/bypass'
        resp, _ = helpers.call_api(url, 'post', json_object=req_body)

        if resp is None:
            logger.debug('Failed to get details for %s', self.device_name)
            return False
        if resp.get('code') == -11300030:
            logger.debug('%s is offline', self.device_name)
            self.fryer_status.set_standby()
            self.fryer_status.cook_status = 'offline'
            return False
        if resp.get('code') != 0:
            logger.debug('Failed to get details for %s \n with code: %s and message: %s',
                         self.device_name, str(resp.get("code", 0)), resp.get("msg", ''))
            return False
        return_status = resp.get('result', {}).get('returnStatus')
        if return_status is None:
            return False
        self.fryer_status.status_response(return_status)
        return True

    @check_status
    def end(self):
        """End the cooking process."""
        if self.preheat is False \
                and self.fryer_status.cook_status in ['cookStop', 'cooking']:
            cmd = {
                'cookMode': {
                    'cookStatus': 'end'
                }
            }
        elif self.preheat is True \
                and self.fryer_status.cook_status in ['preheatStop', 'heating']:
            cmd = {
                'preheat': {
                    'cookStatus': 'end'
                }
            }
        else:
            logger.debug('Cannot end %s as it is not cooking or preheating',
                         self.device_name)
            return False

        if self._status_api(cmd) is True:
            self.fryer_status.set_standby()
            return True
        return False

    @check_status
    def pause(self) -> bool:
        """Pause the cooking process."""
        if self.cook_status not in ['cooking', 'heating']:
            logger.debug('Cannot pause %s as it is not cooking or preheating',
                         self.device_name)
            return False
        if self.preheat is True:
            cmd = {
                'preheat': {
                    'preheatStatus': 'stop'
                }
            }
        else:
            cmd = {
                'cookMode': {
                    'cookStatus': 'stop'
                }
            }
        if self._status_api(cmd) is True:
            if self.preheat is True:
                self.fryer_status.cook_status = 'preheatStop'
            else:
                self.fryer_status.cook_status = 'cookStop'
            return True
        return False

    def _validate_temp(self, set_temp: int) -> bool:
        """Temperature validation."""
        if self.fryer_status.temp_unit == 'fahrenheight':
            if set_temp < 200 or set_temp > 400:
                logger.debug('Invalid temperature %s for %s', set_temp, self.device_name)
                return False
        if self.fryer_status.temp_unit == 'celsius':
            if set_temp < 75 or set_temp > 205:
                logger.debug('Invalid temperature %s for %s', set_temp, self.device_name)
                return False
        return True

    @check_status
    def cook(self, set_temp: int, set_time: int) -> bool:
        """Set cook time and temperature in Minutes."""
        if self._validate_temp(set_temp) is False:
            return False
        return self._set_cook(set_temp, set_time)

    @check_status
    def resume(self) -> bool:
        """Resume paused preheat or cook."""
        if self.cook_status not in ['preheatStop', 'cookStop']:
            logger.debug('Cannot resume %s as it is not paused', self.device_name)
            return False
        if self.preheat is True:
            cmd = {
                'preheat': {
                    'preheatStatus': 'heating'
                }
            }
        else:
            cmd = {
                'cookMode': {
                    'cookStatus': 'cooking'
                }
            }
        if self._status_api(cmd) is True:
            if self.preheat is True:
                self.fryer_status.cook_status = 'heating'
            else:
                self.fryer_status.cook_status = 'cooking'
            return True
        return False

    @check_status
    def set_preheat(self, target_temp: int, cook_time: int) -> bool:
        """Set preheat mode with cooking time."""
        if self.cook_status not in ['standby', 'cookEnd', 'preheatEnd']:
            logger.debug('Cannot set preheat for %s as it is not in standby',
                         self.device_name)
            return False
        if self._validate_temp(target_temp) is False:
            return False
        cmd = self._cmd_api_dict
        cmd['preheatSetTime'] = 5
        cmd['preheatStatus'] = 'heating'
        cmd['targetTemp'] = target_temp
        cmd['cookSetTime'] = cook_time
        json_cmd = {
            'preheat': cmd
        }
        return self._status_api(json_cmd)

    # @check_status
    def cook_from_preheat(self) -> bool:
        """Start Cook when preheat has ended."""
        if self.preheat is False or self.cook_status != 'preheatEnd':
            logger.debug('Cannot start cook from preheat for %s', self.device_name)
            return False
        return self._set_cook(status='cooking')

    def update(self):
        """Update the device details."""
        self.get_details()

    @staticmethod
    def fryer_code_check(code: Union[str, int]) -> Optional[str]:
        """Return the code description."""
        if isinstance(code, str):
            try:
                code = int(code)
            except ValueError:
                return None
        if code == 11903000:
            return 'Error pausing, air fryer is not cooking.'
        if code == 11902000:
            return 'Error setting cook mode, air fryer is already cooking'
        if str(abs(code))[0:5] == 11300:
            return 'Air fryer is offline'
        return None

    @property
    def _cmd_api_base(self) -> dict:
        """Return Base api dictionary for setting status."""
        return {
            "mode": COOK_MODE,
            "accountId": self.manager.account_id,
        }

    @property
    def _cmd_api_dict(self) -> dict:
        """Return API dictionary for setting status."""
        cmd = self._cmd_api_base
        cmd.update({
            "appointmentTs": 0,
            "recipeId": RECIPE_ID,
            "readyStart": self.ready_start,
            "recipeType": RECIPE_TYPE,
            "customRecipe": CUSTOM_RECIPE,
        })
        return cmd

    def _set_cook(self, set_temp: Optional[int] = None,
                  set_time: Optional[int] = None, status: str = 'cooking') -> bool:
        if set_temp is not None and set_time is not None:
            set_cmd = self._cmd_api_dict

            set_cmd['cookSetTime'] = set_time
            set_cmd['cookSetTemp'] = set_temp
        else:
            set_cmd = self._cmd_api_base
        set_cmd['cookStatus'] = status
        cmd = {'cookMode': set_cmd}
        return self._status_api(cmd)

    def _status_api(self, json_cmd: dict):
        """Set API status with jsonCmd."""
        body = self.get_status_body(json_cmd)
        url = '/cloud/v1/deviceManaged/bypass'
        resp, _ = helpers.call_api(url, 'post', json_object=body)
        if resp is None:
            logger.debug('Failed to set status for %s - No response from API',
                         self.device_name)
            return False

        if resp.get('code') != 0 and resp.get('code') is not None:
            debug_msg = self.fryer_code_check(resp.get('code'))
            logger.debug('Failed to set status for %s \n Code: %s and message: %s \n'
                         ' %s', self.device_name, resp.get("code"),
                         resp.get("msg"), debug_msg)
            return False
        self.last_update = int(time.time())
        self.fryer_status.status_request(json_cmd)
        self.update()
        return True

    def displayJSON(self) -> str:
        """Display JSON of device details."""
        sup = super().displayJSON()
        sup_dict = json.loads(sup)
        sup_dict['cook_status'] = self.cook_status
        sup_dict['temp_unit'] = self.temp_unit

        if self.cook_status not in ['standby', 'cookEnd', 'preheatEnd']:
            status_dict = {
                'preheat': self.preheat,
                'current_temp': self.current_temp,
                'cook_set_temp': self.cook_set_temp,
                'cook_set_time': self.cook_set_time,
                'cook_last_time': self.cook_last_time
            }
            if self.preheat is True:
                preheat_dict = {
                    'preheat_last_time': self.preheat_last_time,
                    'preheat_set_time': self.preheat_set_time
                }
                status_dict.update(preheat_dict)
            sup_dict.update(status_dict)
        return json.dumps(sup_dict, indent=4)



"""VeSync API Device Libary."""

import logging
import re
import time
from itertools import chain
from typing import Tuple
from pyvesync.helpers import Helpers
import pyvesync.helpers as helpermodule
from pyvesync.vesyncbasedevice import VeSyncBaseDevice
from pyvesync.vesyncbulb import *   # noqa: F403, F401
import pyvesync.vesyncbulb as bulb_mods
from pyvesync.vesyncfan import *   # noqa: F403, F401
import pyvesync.vesyncfan as fan_mods
from pyvesync.vesyncoutlet import *   # noqa: F403, F401
import pyvesync.vesyncoutlet as outlet_mods
from pyvesync.vesyncswitch import *   # noqa: F403, F401
import pyvesync.vesynckitchen as kitchen_mods
import pyvesync.vesyncswitch as switch_mods

logger = logging.getLogger(__name__)

API_RATE_LIMIT: int = 30
DEFAULT_TZ: str = 'America/New_York'

DEFAULT_ENER_UP_INT: int = 21600


def object_factory(dev_type, config, manager) -> Tuple[str, VeSyncBaseDevice]:
    """Get device type and instantiate class.

    Pulls the device types from each module to determine the type of device and
    instantiates the device object.

    Args:
        dev_type (str): Device model type returned from API
        config (dict): Device configuration from `VeSync.get_devices()` API call
        manager (VeSync): VeSync manager object

    Returns:
        Tuple[str, VeSyncBaseDevice]: Tuple of device type classification and
        instantiated device object

    Note:
        Device types are pulled from the `*_mods` attribute of each device module.
        See [pyvesync.vesyncbulb.bulb_mods], [pyvesync.vesyncfan.fan_mods],
        [pyvesync.vesyncoutlet.outlet_mods], [pyvesync.vesyncswitch.switch_mods],
        and [pyvesync.vesynckitchen.kitchen_mods] for more information.
    """
    def fans(dev_type, config, manager):
        fan_cls = fan_mods.fan_modules[dev_type]  # noqa: F405
        fan_obj = getattr(fan_mods, fan_cls)
        return 'fans', fan_obj(config, manager)

    def outlets(dev_type, config, manager):
        outlet_cls = outlet_mods.outlet_modules[dev_type]  # noqa: F405
        outlet_obj = getattr(outlet_mods, outlet_cls)
        return 'outlets', outlet_obj(config, manager)

    def switches(dev_type, config, manager):
        switch_cls = switch_mods.switch_modules[dev_type]  # noqa: F405
        switch_obj = getattr(switch_mods, switch_cls)
        return 'switches', switch_obj(config, manager)

    def bulbs(dev_type, config, manager):
        bulb_cls = bulb_mods.bulb_modules[dev_type]  # noqa: F405
        bulb_obj = getattr(bulb_mods, bulb_cls)
        return 'bulbs', bulb_obj(config, manager)

    def kitchen(dev_type, config, manager):
        kitchen_cls = kitchen_mods.kitchen_modules[dev_type]
        kitchen_obj = getattr(kitchen_mods, kitchen_cls)
        return 'kitchen', kitchen_obj(config, manager)

    if dev_type in fan_mods.fan_modules:  # type: ignore  # noqa: F405
        type_str, dev_obj = fans(dev_type, config, manager)
    elif dev_type in outlet_mods.outlet_modules:  # type: ignore  # noqa: F405
        type_str, dev_obj = outlets(dev_type, config, manager)
    elif dev_type in switch_mods.switch_modules:  # type: ignore  # noqa: F405
        type_str, dev_obj = switches(dev_type, config, manager)
    elif dev_type in bulb_mods.bulb_modules:  # type: ignore  # noqa: F405
        type_str, dev_obj = bulbs(dev_type, config, manager)
    elif dev_type in kitchen_mods.kitchen_modules:
        type_str, dev_obj = kitchen(dev_type, config, manager)
    else:
        logger.debug('Unknown device named %s model %s',
                     config.get('deviceName', ''),
                     config.get('deviceType', '')
                     )
        type_str = 'unknown'
        dev_obj = None
    return type_str, dev_obj


class VeSync:  # pylint: disable=function-redefined
    """VeSync Manager Class."""

    def __init__(self, username, password, time_zone=DEFAULT_TZ,
                 debug=False, redact=True):
        """Initialize VeSync Manager.

        This class is used as the manager for all VeSync objects, all methods and
        API calls are performed from this class. Time zone, debug and redact are
        optional. Time zone must be a string of an IANA time zone format. Once
        class is instantiated, call `manager.login()` to log in to VeSync servers,
        which returns `True` if successful. Once logged in, call `manager.update()`
        to retrieve devices and update device details.

        Parameters:
            username : str
                VeSync account username (usually email address)
            password : str
                VeSync account password
            time_zone : str, optional
                Time zone for device from IANA database, by default DEFAULT_TZ
            debug : bool, optional
                Enable debug logging, by default False
            redact : bool, optional
                Redact sensitive information in logs, by default True

        Attributes:
            fans : list
                List of VeSyncFan objects for humidifiers and air purifiers
            outlets : list
                List of VeSyncOutlet objects for smart plugs
            switches : list
                List of VeSyncSwitch objects for wall switches
            bulbs : list
                List of VeSyncBulb objects for smart bulbs
            kitchen : list
                List of VeSyncKitchen objects for smart kitchen appliances
            dev_list : dict
                Dictionary of device lists
            token : str
                VeSync API token
            account_id : str
                VeSync account ID
            enabled : bool
                True if logged in to VeSync, False if not
        """
        self.debug = debug
        if debug:  # pragma: no cover
            logger.setLevel(logging.DEBUG)
            bulb_mods.logger.setLevel(logging.DEBUG)
            switch_mods.logger.setLevel(logging.DEBUG)
            outlet_mods.logger.setLevel(logging.DEBUG)
            fan_mods.logger.setLevel(logging.DEBUG)
            helpermodule.logger.setLevel(logging.DEBUG)
            kitchen_mods.logger.setLevel(logging.DEBUG)
        self._redact = redact
        if redact:
            self.redact = redact
        self.username = username
        self.password = password
        self.token = None
        self.account_id = None
        self.country_code = None
        self.devices = None
        self.enabled = False
        self.update_interval = API_RATE_LIMIT
        self.last_update_ts = None
        self.in_process = False
        self._energy_update_interval = DEFAULT_ENER_UP_INT
        self._energy_check = True
        self._dev_list = {}
        self.outlets = []
        self.switches = []
        self.fans = []
        self.bulbs = []
        self.scales = []
        self.kitchen = []

        self._dev_list = {
            'fans': self.fans,
            'outlets': self.outlets,
            'switches': self.switches,
            'bulbs': self.bulbs,
            'kitchen': self.kitchen
        }

        if isinstance(time_zone, str) and time_zone:
            reg_test = r'[^a-zA-Z/_]'
            if bool(re.search(reg_test, time_zone)):
                self.time_zone = DEFAULT_TZ
                logger.debug('Invalid characters in time zone - %s',
                             time_zone)
            else:
                self.time_zone = time_zone
        else:
            self.time_zone = DEFAULT_TZ
            logger.debug('Time zone is not a string')

    @property
    def debug(self) -> bool:
        """Return debug flag."""
        return self._debug

    @debug.setter
    def debug(self, new_flag: bool) -> None:
        """Set debug flag."""
        log_modules = [bulb_mods,
                       switch_mods,
                       outlet_mods,
                       fan_mods,
                       helpermodule]
        if new_flag:
            logger.setLevel(logging.DEBUG)
            for m in log_modules:
                m.logger.setLevel(logging.DEBUG)
        elif new_flag is False:
            logger.setLevel(logging.WARNING)
            for m in log_modules:
                m.logger.setLevel(logging.WARNING)
        self._debug = new_flag

    @property
    def redact(self) -> bool:
        """Return debug flag."""
        return self._redact

    @redact.setter
    def redact(self, new_flag: bool) -> None:
        """Set debug flag."""
        if new_flag:
            Helpers.shouldredact = True
        elif new_flag is False:
            Helpers.shouldredact = False
        self._redact = new_flag

    @property
    def energy_update_interval(self) -> int:
        """Return energy update interval."""
        return self._energy_update_interval

    @energy_update_interval.setter
    def energy_update_interval(self, new_energy_update: int) -> None:
        """Set energy update interval in seconds."""
        if new_energy_update > 0:
            self._energy_update_interval = new_energy_update

    @staticmethod
    def remove_dev_test(device, new_list: list) -> bool:
        """Test if device should be removed - False = Remove."""
        if isinstance(new_list, list) and device.cid:
            for item in new_list:
                device_found = False
                if 'cid' in item:
                    if device.cid == item['cid']:
                        device_found = True
                        break
                else:
                    logger.debug('No cid found in - %s', str(item))
            if not device_found:
                logger.debug(
                    'Device removed - %s - %s',
                    device.device_name, device.device_type
                )
                return False
        return True

    def add_dev_test(self, new_dev: dict) -> bool:
        """Test if new device should be added - True = Add."""
        if 'cid' in new_dev:
            for _, v in self._dev_list.items():
                for dev in v:
                    if (
                        dev.cid == new_dev.get('cid')
                        and new_dev.get('subDeviceNo', 0) == dev.sub_device_no
                    ):
                        return False
        return True

    def remove_old_devices(self, devices: list) -> bool:
        """Remove devices not found in device list return."""
        for k, v in self._dev_list.items():
            before = len(v)
            v[:] = [x for x in v if self.remove_dev_test(x, devices)]
            after = len(v)
            if before != after:
                logger.debug('%s %s removed', str((before - after)), k)
        return True

    @staticmethod
    def set_dev_id(devices: list) -> list:
        """Correct devices without cid or uuid."""
        dev_num = 0
        dev_rem = []
        for dev in devices:
            if dev.get('cid') is None:
                if dev.get('macID') is not None:
                    dev['cid'] = dev['macID']
                elif dev.get('uuid') is not None:
                    dev['cid'] = dev['uuid']
                else:
                    dev_rem.append(dev_num)
                    logger.warning('Device with no ID  - %s',
                                   dev.get('deviceName'))
            dev_num += 1
            if dev_rem:
                devices = [i for j, i in enumerate(
                            devices) if j not in dev_rem]
        return devices

    def process_devices(self, dev_list: list) -> bool:
        """Instantiate Device Objects.

        Internal method run by `get_devices()` to instantiate device objects.

        """
        devices = VeSync.set_dev_id(dev_list)

        num_devices = 0
        for _, v in self._dev_list.items():
            if isinstance(v, list):
                num_devices += len(v)
            else:
                num_devices += 1

        if not devices:
            logger.warning('No devices found in api return')
            return False
        if num_devices == 0:
            logger.debug('New device list initialized')
        else:
            self.remove_old_devices(devices)

        devices[:] = [x for x in devices if self.add_dev_test(x)]

        detail_keys = ['deviceType', 'deviceName', 'deviceStatus']
        for dev in devices:
            if not all(k in dev for k in detail_keys):
                logger.debug('Error adding device')
                continue
            dev_type = dev.get('deviceType')
            try:
                device_str, device_obj = object_factory(dev_type, dev, self)
                device_list = getattr(self, device_str)
                device_list.append(device_obj)
            except AttributeError as err:
                logger.debug('Error - %s', err)
                logger.debug('%s device not added', dev_type)
                continue

        return True

    def get_devices(self) -> bool:
        """Return tuple listing outlets, switches, and fans of devices.

        This is an internal method called by `update()`
        """
        if not self.enabled:
            return False

        self.in_process = True
        proc_return = False
        response, _ = Helpers.call_api(
            '/cloud/v1/deviceManaged/devices',
            'post',
            headers=Helpers.req_header_bypass(),
            json_object=Helpers.req_body(self, 'devicelist'),
        )

        if response and Helpers.code_check(response):
            if 'result' in response and 'list' in response['result']:
                device_list = response['result']['list']
                proc_return = self.process_devices(device_list)
            else:
                logger.error('Device list in response not found')
        else:
            logger.warning('Error retrieving device list')

        self.in_process = False

        return proc_return

    def login(self) -> bool:
        """Log into VeSync server.

        Username and password are provided when class is instantiated.

        Returns:
            True if login successful, False if not
        """
        user_check = isinstance(self.username, str) and len(self.username) > 0
        pass_check = isinstance(self.password, str) and len(self.password) > 0
        if user_check is False:
            logger.error('Username invalid')
            return False
        if pass_check is False:
            logger.error('Password invalid')
            return False

        response, _ = Helpers.call_api(
            '/cloud/v1/user/login', 'post',
            json_object=Helpers.req_body(self, 'login')
        )

        if Helpers.code_check(response) and 'result' in response:
            self.token = response.get('result').get('token')
            self.account_id = response.get('result').get('accountID')
            self.country_code = response.get('result').get('countryCode')
            self.enabled = True
            logger.debug('Login successful')
            logger.debug('token %s', self.token)
            logger.debug('account_id %s', self.account_id)

            return True
        logger.error('Error logging in with username and password')
        return False

    def device_time_check(self) -> bool:
        """Test if update interval has been exceeded."""
        return (
            self.last_update_ts is None
            or (time.time() - self.last_update_ts) > self.update_interval
        )

    def update(self) -> None:
        """Fetch updated information about devices.

        Pulls devices list from VeSync and instantiates any new devices. Devices
        are stored in the instance attributes `outlets`, `switches`, `fans`, and
        `bulbs`. The `_device_list` attribute is a dictionary of these attributes.
        """
        if self.device_time_check():

            if not self.enabled:
                logger.error('Not logged in to VeSync')
                return
            self.get_devices()

            devices = list(self._dev_list.values())

            logger.debug('Start updating the device details one by one')
            for device in chain(*devices):
                device.update()

            self.last_update_ts = time.time()

    def update_energy(self, bypass_check=False) -> None:
        """Fetch updated energy information for outlet devices."""
        if self.outlets:
            for outlet in self.outlets:
                outlet.update_energy(bypass_check)

    def update_all_devices(self) -> None:
        """Run `get_details()` for each device and update state."""
        devices = list(self._dev_list.values())
        for dev in chain(*devices):
            dev.update()



"""Common base class for tests and Default values.

Routine Listings
----------------

FunctionResponses : defaultdict
    Defaultdict of the standard response tuple for device methods
Defaults: class
    Default values and methods for generating default values
TestBase: class
    Base class for tests to start mock & instantiat VS object
"""
import logging
from pathlib import Path
from typing import Any
import pytest
import yaml
from collections import defaultdict, namedtuple
from unittest.mock import patch
from requests.structures import CaseInsensitiveDict
from pyvesync.vesync import VeSync
from pyvesync.helpers import Color
import pyvesync.helpers as vs_helpers


logger = logging.getLogger(__name__)

FunctionResponses: defaultdict = defaultdict(lambda: ({"code": 0, "msg": None}, 200))

CALL_API_ARGS = ['url', 'method', 'data', 'headers']

ID_KEYS = ['CID', 'UUID', 'MACID']

class Defaults:
    """General defaults for API responses and requests.

    Attributes
    ----------
    token : str
        Default token for API requests
    active_time : str
        Default active time for API responses
    account_id : str
        Default account ID for API responses
    active_time : str
        Default active time for API responses
    color: Color (dataclass)
        Red=50, Green=100, Blue=225, Hue=223, Saturation=77.78, value=88.24
        Default Color dataclass
        contains red, green, blue, hue, saturation and value attributes

    Methods
    --------
    name(dev_type='NA')
        Default device name created from "dev_type-NAME"
    cid(dev_type='NA')
        Default device cid created from "dev_type-CID"
    uuid(dev_type='NA')
        Default device uuid created from "dev_type-UUID"
    macid(dev_type='NA')
        Default device macid created from "dev_type-MACID"
    """

    token = 'sample_tk'
    account_id = 'sample_id'
    trace_id = "TRACE_ID"
    active_time = 1
    color = Color(red=50, green=100, blue=225)
    brightness = 100
    color_temp = 100
    bool_toggle = True
    str_toggle = 'on'
    bin_toggle = 1

    @staticmethod
    def name(dev_type: str = 'NA'):
        """Name of device with format f"{dev_type}-NAME".

        Parameters
        ----------
        dev_type : str
            Device type use to create default name

        Returns
        -------
        str
            Default name for device f"dev_type-NAME"
        """
        return f'{dev_type}-NAME'

    @staticmethod
    def cid(dev_type='NA'):
        """CID for a device with format f"{dev_type}-CID".

        Parameters
        ----------
        dev_type : str
            Device type use to create default cid

        Returns
        -------
        str
            Default cid for device f"dev_type-CID"
        """
        return f'{dev_type}-CID'

    @staticmethod
    def uuid(dev_type='NA'):
        """UUID for a device with format f"{dev_type}-UUID".

        Parameters
        ----------
        dev_type : str
            Device type use to create default UUID

        Returns
        -------
        str
            Default uuid for device f"{dev_type}-UUID"
        """
        return f'{dev_type}-UUID'

    @staticmethod
    def macid(dev_type='NA'):
        """MACID for a device with format f"{dev_type}-MACID".

        Parameters
        ----------
        dev_type : str
            Device type use to create default macid

        Returns
        -------
        str
            Default macID for device f"{dev_type}-MACID"
        """
        return f'{dev_type}-MACID'


def parse_args(mock_api):
    """Parse arguments from mock API call.

    Arguments
    ----------
    mock_api : mock
        Mock object used to path call_api() method

    Returns
    -------
    dict
        dictionary of all call_api() arguments
    """
    call_args = mock_api.call_args.args
    call_kwargs = mock_api.call_args.kwargs
    all_kwargs = dict(zip(CALL_API_ARGS, call_args))
    all_kwargs.update(call_kwargs)
    return all_kwargs


API_DEFAULTS = CaseInsensitiveDict({
    'accountID': Defaults.account_id,
    'token': Defaults.token,
    'timeZone': vs_helpers.DEFAULT_TZ,
    'acceptLanguage': 'en',
    'appVersion': vs_helpers.APP_VERSION,
    'phoneBrand': vs_helpers.PHONE_BRAND,
    'phoneOS': vs_helpers.PHONE_OS,
    'userType': vs_helpers.USER_TYPE,
    "tk": Defaults.token,
    "traceId": "TRACE_ID",
    'verifyEmail': 'EMAIL',
    'nickName': 'NICKNAME',
    'password': 'PASSWORD',
    'username': 'EMAIL',
    'email': 'EMAIL',
    'deviceName': 'NAME'

})


class YAMLWriter:
    """Read and Write API request data to YAML files.

    Arguments
    ---------
    module : str
        name of module that is being tested
    dev_type : str
        device type being tested

    Attributes
    ----------
    self.file_path : Path
        Path to YAML directory, default to API dir in tests folder
    self.file : Path
        Path to YAML file based on device type and module
    self.existings_yaml : dict
        Existing YAML data read to dict object
    self._existing_api : dict, optional
        Existing data dict of a specific API call

    Methods
    -------
    self.existing_api()
        Return existing data dict of a specific API call or None
    self.write_api(api, data, overwrite=False)
        Writes data to YAML file for a specific API call,
        set overwrite=True to overwrite existing data

    """

    def __init__(self, module, dev_type):
        """Init the YAMLWriter class.

        Arguments
        ----------
        module : str
            name of module that is being tested
        dev_type : str
            device type being tested
        """
        self.file_path = self._get_path(module)
        self.file = Path.joinpath(self.file_path, dev_type + '.yaml')
        self.new_file = self._new_file()
        self.existing_yaml = self._get_existing_yaml()
        self._existing_api = None

    @staticmethod
    def _get_path(module) -> Path:
        yaml_dir = Path.joinpath(Path(__file__).parent, 'api', module)
        if not yaml_dir.exists():
            yaml_dir.mkdir(parents=True)
        return yaml_dir

    def _new_file(self) -> bool:
        if not self.file.exists():
            logger.debug(f'Creating new file {self.file}')
            self.file.touch()
            return True
        return False

    def _get_existing_yaml(self) -> Any:
        if self.new_file:
            return None
        with open(self.file, 'rb') as f:
            data = yaml.full_load(f)
        return data

    def existing_api(self, method) -> bool:
        """Check YAML file for existing data for API call.

        Arguments
        ----------
        method : str
            Name of method being tested

        Returns
        -------
        dict or None
            Existing data for API call or None
        """
        if self.existing_yaml is not None:
            current_dict = self.existing_yaml.get(method)
            self._existing_api = current_dict
            if current_dict is not None:
                logger.debug(f'API call {method} already exists in {self.file}')
                return True
        return False

    def write_api(self, method, yaml_dict, overwrite=False):
        """Write API data to YAML file.

        Arguments
        ----------
        method : str
            Name of method being tested
        yaml_dict : dict
            Data to write to YAML file
        overwrite : bool, optional
            Overwrite existing data, default to False
        """
        if self.existing_yaml is not None:
            current_dict = self.existing_yaml.get(method)
            if current_dict is not None and overwrite is False:
                logger.debug(f'API call {method} already exists in {self.file}')
                return
            self.existing_yaml[method] = yaml_dict
        else:
            self.existing_yaml = {method: yaml_dict}
        with open(self.file, 'w', encoding='utf-8') as f:
            yaml.dump(self.existing_yaml, f, encoding='utf-8')


def api_scrub(api_dict, device_type=None):
    """Recursive function to scrub all sensitive data from API call.

    Arguments
    ----------
    api_dict : dict
        API call data to scrub
    device_type : str, optional
        Device type to use for default values

    Returns
    -------
    dict
        Scrubbed API call data
    """
    def id_cleaner(key, value):
        if key.upper() in ID_KEYS:
            return f"{device_type or 'UNKNOWN'}-{key.upper()}"
        if key in API_DEFAULTS:
            return API_DEFAULTS[key]
        return value

    def nested_dict_iter(nested, mapper, last_key=None):
        if isinstance(nested, dict):
            if nested.get('deviceType') is not None:
                nonlocal device_type
                device_type = nested['deviceType']
            out = {}
            for key, value in nested.items():
                out[key] = nested_dict_iter(value, mapper, key)
            return out

        if isinstance(nested, list):
            return [nested_dict_iter(el, mapper, last_key) for el in nested]
        if not last_key:
            return nested
        return mapper(last_key, nested)
    return nested_dict_iter(api_dict, id_cleaner)


class TestBase:
    """Base class for all tests.

    Contains instantiated VeSync object and mocked
    API call for call_api() function.

    Attributes
    ----------
    self.mock_api : Mock
        Mock for call_api() function
    self.manager : VeSync
        Instantiated VeSync object that is logged in
    self.caplog : LogCaptureFixture
        Pytest fixture for capturing logs
    """
    overwrite = False
    write_api = False

    @pytest.fixture(autouse=True, scope='function')
    def setup(self, caplog):
        """Fixture to instantiate VeSync object, start logging and start Mock.

        Attributes
        ----------
        self.mock_api : Mock
        self.manager : VeSync
        self.caplog : LogCaptureFixture

        Yields
        ------
        Class instance with mocked call_api() function and VeSync object
        """
        self.mock_api_call = patch('pyvesync.helpers.Helpers.call_api')
        self.caplog = caplog
        self.mock_api = self.mock_api_call.start()
        self.mock_api.create_autospect()
        self.mock_api.return_value.ok = True
        self.manager = VeSync('EMAIL', 'PASSWORD', debug=True)
        self.manager.enabled = True
        self.manager.token = Defaults.token
        self.manager.account_id = Defaults.account_id
        caplog.set_level(logging.DEBUG)
        yield
        self.mock_api_call.stop()


def assert_test(test_func, all_kwargs, dev_type=None,
                 write_api=False, overwrite=False):
    """Test pyvesync API calls against existing API.

    Set `write_api=True` to True to write API call data to YAML file. 
    This will not overwrite existing data unless overwrite is True.
    The overwrite argument is only used when API changes, defaults
    to false for development testing. `overwrite=True` and `write_api=True`
    need to both be set to overwrite existing data.

    Arguments
    ----------
    test_func : method
        Method that is being tested
    all_kwargs : dict
        Dictionary of call_api() arguments
    dev_type : str, optional
        Device type being tested
    write_api : bool, optional
        Write API call data to YAML file, default to False
    overwrite : bool, optional
        Overwrite existing data ONLY USE FOR CHANGING API,
          default to False. Must be set with `write_api=True`

    Returns
    -------
    bool
        True if test passes, False if test fails

    Asserts
    -------
        Asserts that the API call data matches the expected data
    """
    if all_kwargs.get('json_object') is not None:
        all_kwargs['json_object'] = api_scrub(all_kwargs['json_object'], dev_type)
    if all_kwargs.get('headers') is not None:
        all_kwargs['headers'] = api_scrub(all_kwargs['headers'], dev_type)
    mod = test_func.__module__.split(".")[-1]
    if dev_type is None:
        cls_name = test_func.__self__.__class__.__name__
    else:
        cls_name = dev_type
    method_name = test_func.__name__
    writer = YAMLWriter(mod, cls_name)
    if overwrite is True and write_api is True:
        writer.write_api(method_name, all_kwargs, overwrite)
        return True
    if writer.existing_api(method_name) is False:
        logger.debug("No existing, API data for %s %s %s", mod, cls_name, method_name)
        if write_api is True:
            logger.debug("Writing API data for %s %s %s", mod, cls_name, method_name)
            writer.write_api(method_name, all_kwargs, overwrite)
        else:
            logger.debug("Not writing API data for %s %s %s", mod, cls_name, method_name)
    assert writer._existing_api == all_kwargs
    return True



"""Etekcity Outlets."""

import logging
import time
import json
from abc import ABCMeta, abstractmethod

from pyvesync.helpers import Helpers
from pyvesync.vesyncbasedevice import VeSyncBaseDevice

logger = logging.getLogger(__name__)

outlet_config = {
    'wifi-switch-1.3': {
        'module': 'VeSyncOutlet7A'},
    'ESW03-USA': {
        'module': 'VeSyncOutlet10A'},
    'ESW01-EU': {
        'module': 'VeSyncOutlet10A'},
    'ESW15-USA': {
        'module': 'VeSyncOutlet15A'},
    'ESO15-TB': {
        'module': 'VeSyncOutdoorPlug'},
    'BSDOG01': {
        'module': 'VeSyncOutletBSDGO1'},
}

outlet_modules = {k: v['module'] for k, v in outlet_config.items()}

__all__ = list(outlet_modules.values()) + ['outlet_modules']


class VeSyncOutlet(VeSyncBaseDevice):
    """Base class for Etekcity Outlets."""

    __metaclass__ = ABCMeta

    def __init__(self, details, manager):
        """Initilize VeSync Outlet base class."""
        super().__init__(details, manager)

        self.details = {}
        self.energy = {}
        self.update_energy_ts = None
        self._energy_update_interval = manager.energy_update_interval

    @property
    def update_time_check(self) -> bool:
        """Test if energy update interval has been exceeded."""
        if self.update_energy_ts is None:
            return True

        if ((time.time() - self.update_energy_ts)
                > self._energy_update_interval):
            return True
        return False

    @abstractmethod
    def turn_on(self) -> bool:
        """Return True if device has beeeen turned on."""

    @abstractmethod
    def turn_off(self) -> bool:
        """Return True if device has beeeen turned off."""

    @abstractmethod
    def get_details(self) -> None:
        """Build details dictionary."""

    @abstractmethod
    def get_weekly_energy(self) -> None:
        """Build weekly energy history dictionary."""

    @abstractmethod
    def get_monthly_energy(self) -> None:
        """Build Monthly Energy History Dictionary."""

    @abstractmethod
    def get_yearly_energy(self):
        """Build Yearly Energy Dictionary."""

    @abstractmethod
    def get_config(self):
        """Get configuration and firmware details."""

    def update(self):
        """Get Device Energy and Status."""
        self.get_details()

    def update_energy(self, bypass_check: bool = False):
        """Build weekly, monthly and yearly dictionaries."""
        if bypass_check or (not bypass_check and self.update_time_check):
            self.update_energy_ts = time.time()
            self.get_weekly_energy()
            if 'week' in self.energy:
                self.get_monthly_energy()
                self.get_yearly_energy()
            if not bypass_check:
                self.update_energy_ts = time.time()

    @property
    def active_time(self) -> int:
        """Return active time of a device in minutes."""
        return self.details.get('active_time', 0)

    @property
    def energy_today(self) -> float:
        """Return energy."""
        return self.details.get('energy', 0)

    @property
    def power(self) -> float:
        """Return current power in watts."""
        return float(self.details.get('power', 0))

    @property
    def voltage(self) -> float:
        """Return current voltage."""
        return float(self.details.get('voltage', 0))

    @property
    def monthly_energy_total(self) -> float:
        """Return total energy usage over the month."""
        return self.energy.get('month', {}).get('total_energy', 0)

    @property
    def weekly_energy_total(self) -> float:
        """Return total energy usage over the week."""
        return self.energy.get('week', {}).get('total_energy', 0)

    @property
    def yearly_energy_total(self) -> float:
        """Return total energy usage over the year."""
        return self.energy.get('year', {}).get('total_energy', 0)

    def display(self):
        """Return formatted device info to stdout."""
        super().display()
        disp = [
            ('Active Time : ', self.active_time, ' minutes'),
            ('Energy: ', self.energy_today, ' kWh'),
            ('Power: ', self.power, ' Watts'),
            ('Voltage: ', self.voltage, ' Volts'),
            ('Energy Week: ', self.weekly_energy_total, ' kWh'),
            ('Energy Month: ', self.monthly_energy_total, ' kWh'),
            ('Energy Year: ', self.yearly_energy_total, ' kWh'),
        ]
        for line in disp:
            print(f'{line[0]:.<30} {line[1]} {line[2]}')

    def displayJSON(self):
        """Return JSON details for outlet."""
        sup = super().displayJSON()
        sup_val = json.loads(sup)
        sup_val.update(
            {
                'Active Time': str(self.active_time),
                'Energy': str(self.energy_today),
                'Power': str(self.power),
                'Voltage': str(self.voltage),
                'Energy Week': str(self.weekly_energy_total),
                'Energy Month': str(self.monthly_energy_total),
                'Energy Year': str(self.yearly_energy_total),
            }
        )

        return json.dumps(sup_val, indent=4)


class VeSyncOutlet7A(VeSyncOutlet):
    """Etekcity 7A Round Outlet Class."""

    def __init__(self, details, manager):
        """Initilize Etekcity 7A round outlet class."""
        super().__init__(details, manager)
        self.det_keys = ['deviceStatus', 'activeTime',
                         'energy', 'power', 'voltage']
        self.energy_keys = ['energyConsumptionOfToday',
                            'maxEnergy', 'totalEnergy']

    def get_details(self) -> None:
        """Get 7A outlet details."""
        r, _ = Helpers.call_api(
            '/v1/device/' + self.cid + '/detail',
            'get',
            headers=Helpers.req_headers(self.manager),
        )

        if r is not None and all(x in r for x in self.det_keys):
            self.device_status = r.get('deviceStatus', self.device_status)
            self.details['active_time'] = r.get('activeTime', 0)
            self.details['energy'] = r.get('energy', 0)
            power = r.get('power', '0')
            self.details['power'] = self.parse_energy_detail(power)
            voltage = r.get('voltage', 0)
            self.details['voltage'] = self.parse_energy_detail(voltage)
        else:
            logger.debug('Unable to get %s details', self.device_name)

    @staticmethod
    def parse_energy_detail(energy):
        """Parse energy details to be compatible with new and old firmware."""
        try:
            if isinstance(energy, str) and ':' in energy:
                power = round(float(Helpers.calculate_hex(energy)), 2)
            else:
                power = float(energy)
        except ValueError:
            logger.debug('Error parsing power response - %s', energy)
            power = 0
        return power

    def get_weekly_energy(self) -> None:
        """Get 7A outlet weekly energy info and buld weekly energy dict."""
        r, _ = Helpers.call_api(
            '/v1/device/' + self.cid + '/energy/week',
            'get',
            headers=Helpers.req_headers(self.manager),
        )

        if r is not None and all(x in r for x in self.energy_keys):
            self.energy['week'] = Helpers.build_energy_dict(r)
        else:
            logger.debug('Unable to get %s weekly data', self.device_name)

    def get_monthly_energy(self) -> None:
        """Get 7A outlet monthly energy info and buld monthly energy dict."""
        r, _ = Helpers.call_api(
            '/v1/device/' + self.cid + '/energy/month',
            'get',
            headers=Helpers.req_headers(self.manager),
        )

        if r is not None and all(x in r for x in self.energy_keys):
            self.energy['month'] = Helpers.build_energy_dict(r)
        else:
            logger.warning('Unable to get %s monthly data', self.device_name)

    def get_yearly_energy(self) -> None:
        """Get 7A outlet yearly energy info and build yearly energy dict."""
        r, _ = Helpers.call_api(
            '/v1/device/' + self.cid + '/energy/year',
            'get',
            headers=Helpers.req_headers(self.manager),
        )

        if r is not None and all(x in r for x in self.energy_keys):
            self.energy['year'] = Helpers.build_energy_dict(r)
        else:
            logger.debug('Unable to get %s yearly data', self.device_name)

    def turn_on(self) -> bool:
        """Turn 7A outlet on - return True if successful."""
        _, status_code = Helpers.call_api(
            '/v1/wifi-switch-1.3/' + self.cid + '/status/on',
            'put',
            headers=Helpers.req_headers(self.manager),
        )

        if status_code is not None and status_code == 200:
            self.device_status = 'on'

            return True
        logger.warning('Error turning %s on', self.device_name)
        return False

    def turn_off(self) -> bool:
        """Turn 7A outlet off - return True if successful."""
        _, status_code = Helpers.call_api(
            '/v1/wifi-switch-1.3/' + self.cid + '/status/off',
            'put',
            headers=Helpers.req_headers(self.manager),
        )

        if status_code is not None and status_code == 200:
            self.device_status = 'off'

            return True
        logger.warning('Error turning %s off', self.device_name)
        return False

    def get_config(self) -> None:
        """Get 7A outlet configuration info."""
        r, _ = Helpers.call_api(
            '/v1/device/' + self.cid + '/configurations',
            'get',
            headers=Helpers.req_headers(self.manager),
        )

        if 'currentFirmVersion' in r:
            self.config = Helpers.build_config_dict(r)
        else:
            logger.debug('Error getting configuration info for %s',
                         self.device_name)


class VeSyncOutlet10A(VeSyncOutlet):
    """Etekcity 10A Round Outlets."""

    def __init__(self, details, manager):
        """Initialize 10A outlet class."""
        super().__init__(details, manager)

    def get_details(self) -> None:
        """Get 10A outlet details."""
        body = Helpers.req_body(self.manager, 'devicedetail')
        body['uuid'] = self.uuid

        r, _ = Helpers.call_api(
            '/10a/v1/device/devicedetail',
            'post',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(r):
            self.device_status = r.get('deviceStatus', self.device_status)
            self.connection_status = r.get('connectionStatus',
                                           self.connection_status)
            self.details = Helpers.build_details_dict(r)
        else:
            logger.debug('Unable to get %s details', self.device_name)

    def get_config(self) -> None:
        """Get 10A outlet configuration info."""
        body = Helpers.req_body(self.manager, 'devicedetail')
        body['method'] = 'configurations'
        body['uuid'] = self.uuid

        r, _ = Helpers.call_api(
            '/10a/v1/device/configurations',
            'post',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(r):
            self.config = Helpers.build_config_dict(r)
        else:
            logger.debug('Error getting %s config info', self.device_name)

    def get_weekly_energy(self) -> None:
        """Get 10A outlet weekly energy info and populate energy dict."""
        body = Helpers.req_body(self.manager, 'energy_week')
        body['uuid'] = self.uuid

        response, _ = Helpers.call_api(
            '/10a/v1/device/energyweek',
            'post',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(response):
            self.energy['week'] = Helpers.build_energy_dict(response)
        else:
            logger.debug('Unable to get %s weekly data', self.device_name)

    def get_monthly_energy(self) -> None:
        """Get 10A outlet monthly energy info and populate energy dict."""
        body = Helpers.req_body(self.manager, 'energy_month')
        body['uuid'] = self.uuid

        response, _ = Helpers.call_api(
            '/10a/v1/device/energymonth',
            'post',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(response):
            self.energy['month'] = Helpers.build_energy_dict(response)
        else:
            logger.debug('Unable to get %s monthly data', self.device_name)

    def get_yearly_energy(self) -> None:
        """Get 10A outlet yearly energy info and populate energy dict."""
        body = Helpers.req_body(self.manager, 'energy_year')
        body['uuid'] = self.uuid

        response, _ = Helpers.call_api(
            '/10a/v1/device/energyyear',
            'post',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(response):
            self.energy['year'] = Helpers.build_energy_dict(response)
        else:
            logger.debug('Unable to get %s yearly data', self.device_name)

    def turn_on(self) -> bool:
        """Turn 10A outlet on - return True if successful."""
        body = Helpers.req_body(self.manager, 'devicestatus')
        body['uuid'] = self.uuid
        body['status'] = 'on'

        response, _ = Helpers.call_api(
            '/10a/v1/device/devicestatus',
            'put',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(response):
            self.device_status = 'on'
            return True
        logger.warning('Error turning %s on', self.device_name)
        return False

    def turn_off(self) -> bool:
        """Turn 10A outlet off - return True if successful."""
        body = Helpers.req_body(self.manager, 'devicestatus')
        body['uuid'] = self.uuid
        body['status'] = 'off'

        response, _ = Helpers.call_api(
            '/10a/v1/device/devicestatus',
            'put',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(response):
            self.device_status = 'off'
            return True
        logger.warning('Error turning %s off', self.device_name)
        return False


class VeSyncOutlet15A(VeSyncOutlet):
    """Class for Etekcity 15A Rectangular Outlets."""

    def __init__(self, details, manager):
        """Initialize 15A rectangular outlets."""
        super().__init__(details, manager)
        self.nightlight_status = 'off'
        self.nightlight_brightness = 0

    def get_details(self) -> None:
        """Get 15A outlet details."""
        body = Helpers.req_body(self.manager, 'devicedetail')
        body['uuid'] = self.uuid

        r, _ = Helpers.call_api(
            '/15a/v1/device/devicedetail',
            'post',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        attr_list = (
            'deviceStatus',
            'activeTime',
            'energy',
            'power',
            'voltage',
            'nightLightStatus',
            'nightLightAutomode',
            'nightLightBrightness',
        )

        if Helpers.code_check(r) and all(k in r for k in attr_list):

            self.device_status = r.get('deviceStatus')
            self.connection_status = r.get('connectionStatus')
            self.nightlight_status = r.get('nightLightStatus')
            self.nightlight_brightness = r.get('nightLightBrightness')
            self.details = Helpers.build_details_dict(r)
        else:
            logger.debug('Unable to get %s details', self.device_name)

    def get_config(self) -> None:
        """Get 15A outlet configuration info."""
        body = Helpers.req_body(self.manager, 'devicedetail')
        body['method'] = 'configurations'
        body['uuid'] = self.uuid

        r, _ = Helpers.call_api(
            '/15a/v1/device/configurations',
            'post',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(r):
            self.config = Helpers.build_config_dict(r)
        else:
            logger.debug('Unable to get %s config info', self.device_name)

    def get_weekly_energy(self) -> None:
        """Get 15A outlet weekly energy info and populate energy dict."""
        body = Helpers.req_body(self.manager, 'energy_week')
        body['uuid'] = self.uuid

        response, _ = Helpers.call_api(
            '/15a/v1/device/energyweek',
            'post',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(response):
            self.energy['week'] = Helpers.build_energy_dict(response)
        else:
            logger.debug('Unable to get %s weekly data', self.device_name)

    def get_monthly_energy(self) -> None:
        """Get 15A outlet monthly energy info and populate energy dict."""
        body = Helpers.req_body(self.manager, 'energy_month')
        body['uuid'] = self.uuid

        response, _ = Helpers.call_api(
            '/15a/v1/device/energymonth',
            'post',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(response):
            self.energy['month'] = Helpers.build_energy_dict(response)
        else:
            logger.debug('Unable to get %s monthly data', self.device_name)

    def get_yearly_energy(self) -> None:
        """Get 15A outlet yearly energy info and populate energy dict."""
        body = Helpers.req_body(self.manager, 'energy_year')
        body['uuid'] = self.uuid

        response, _ = Helpers.call_api(
            '/15a/v1/device/energyyear',
            'post',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(response):
            self.energy['year'] = Helpers.build_energy_dict(response)
        else:
            logger.debug('Unable to get %s yearly data', self.device_name)

    def turn_on(self) -> bool:
        """Turn 15A outlet on - return True if successful."""
        body = Helpers.req_body(self.manager, 'devicestatus')
        body['uuid'] = self.uuid
        body['status'] = 'on'

        response, _ = Helpers.call_api(
            '/15a/v1/device/devicestatus',
            'put',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(response):
            self.device_status = 'on'
            return True
        logger.warning('Error turning %s on', self.device_name)
        return False

    def turn_off(self) -> bool:
        """Turn 15A outlet off - return True if successful."""
        body = Helpers.req_body(self.manager, 'devicestatus')
        body['uuid'] = self.uuid
        body['status'] = 'off'

        response, _ = Helpers.call_api(
            '/15a/v1/device/devicestatus',
            'put',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(response):
            self.device_status = 'off'
            return True
        logger.warning('Error turning %s off', self.device_name)
        return False

    def turn_on_nightlight(self) -> bool:
        """Turn on nightlight."""
        body = Helpers.req_body(self.manager, 'devicestatus')
        body['uuid'] = self.uuid
        body['mode'] = 'auto'

        response, _ = Helpers.call_api(
            '/15a/v1/device/nightlightstatus',
            'put',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(response):
            return True
        logger.debug('Error turning on %s nightlight', self.device_name)
        return False

    def turn_off_nightlight(self) -> bool:
        """Turn Off Nightlight."""
        body = Helpers.req_body(self.manager, 'devicestatus')
        body['uuid'] = self.uuid
        body['mode'] = 'manual'

        response, _ = Helpers.call_api(
            '/15a/v1/device/nightlightstatus',
            'put',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(response):
            return True
        logger.debug('Error turning off %s nightlight', self.device_name)
        return False


class VeSyncOutdoorPlug(VeSyncOutlet):
    """Class to hold Etekcity outdoor outlets."""

    def __init__(self, details, manager):
        """Initialize Etekcity Outdoor Plug class."""
        super().__init__(details, manager)

    def get_details(self) -> None:
        """Get details for outdoor outlet."""
        body = Helpers.req_body(self.manager, 'devicedetail')
        body['uuid'] = self.uuid
        r, _ = Helpers.call_api(
            '/outdoorsocket15a/v1/device/devicedetail',
            'post',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(r):
            self.details = Helpers.build_details_dict(r)
            self.connection_status = r.get('connectionStatus')

            dev_no = self.sub_device_no
            sub_device_list = r.get('subDevices')
            if sub_device_list and dev_no <= len(sub_device_list):
                self.device_status = sub_device_list[(dev_no + -1)].get(
                    'subDeviceStatus'
                )
                return
        logger.debug('Unable to get %s details', self.device_name)

    def get_config(self) -> None:
        """Get configuration info for outdoor outlet."""
        body = Helpers.req_body(self.manager, 'devicedetail')
        body['method'] = 'configurations'
        body['uuid'] = self.uuid

        r, _ = Helpers.call_api(
            '/outdoorsocket15a/v1/device/configurations',
            'post',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(r):
            self.config = Helpers.build_config_dict(r)
        logger.debug('Error getting %s config info', self.device_name)

    def get_weekly_energy(self) -> None:
        """Get outdoor outlet weekly energy info and populate energy dict."""
        body = Helpers.req_body(self.manager, 'energy_week')
        body['uuid'] = self.uuid

        response, _ = Helpers.call_api(
            '/outdoorsocket15a/v1/device/energyweek',
            'post',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(response):
            self.energy['week'] = Helpers.build_energy_dict(response)
        else:
            logger.debug('Unable to get %s weekly data', self.device_name)

    def get_monthly_energy(self) -> None:
        """Get outdoor outlet monthly energy info and populate energy dict."""
        body = Helpers.req_body(self.manager, 'energy_month')
        body['uuid'] = self.uuid

        response, _ = Helpers.call_api(
            '/outdoorsocket15a/v1/device/energymonth',
            'post',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(response):
            self.energy['month'] = Helpers.build_energy_dict(response)
        logger.debug('Unable to get %s monthly data', self.device_name)

    def get_yearly_energy(self) -> None:
        """Get outdoor outlet yearly energy info and populate energy dict."""
        body = Helpers.req_body(self.manager, 'energy_year')
        body['uuid'] = self.uuid

        response, _ = Helpers.call_api(
            '/outdoorsocket15a/v1/device/energyyear',
            'post',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(response):
            self.energy['year'] = Helpers.build_energy_dict(response)
        else:
            logger.debug('Unable to get %s yearly data', self.device_name)

    def toggle(self, status) -> bool:
        """Toggle power for outdoor outlet."""
        body = Helpers.req_body(self.manager, 'devicestatus')
        body['uuid'] = self.uuid
        body['status'] = status
        body['switchNo'] = self.sub_device_no

        response, _ = Helpers.call_api(
            '/outdoorsocket15a/v1/device/devicestatus',
            'put',
            headers=Helpers.req_headers(self.manager),
            json_object=body,
        )

        if Helpers.code_check(response):
            self.device_status = status
            return True
        logger.warning('Error turning %s %s', self.device_name, status)
        return False

    def turn_on(self) -> bool:
        """Turn on outlet."""
        return bool(self.toggle('on'))

    def turn_off(self) -> bool:
        """Turn off outlet."""
        return bool(self.toggle('off'))


class VeSyncOutletBSDGO1(VeSyncOutlet):
    """VeSync BSDGO1 smart plug."""

    def __init__(self, details, manager):
        """Initialize BSDGO1 smart plug class."""
        super().__init__(details, manager)

    def get_details(self) -> None:
        """Get BSDGO1 device details."""
        body = Helpers.req_body(self.manager, 'bypassV2')
        body['cid'] = self.cid
        body['configModule'] = self.config_module
        body['payload'] = {
            'method': 'getProperty',
            'source': 'APP',
            'data': {}
        }

        r, _ = Helpers.call_api(
            '/cloud/v2/deviceManaged/bypassV2',
            'post',
            headers=Helpers.req_header_bypass(),
            json_object=body,
        )

        if Helpers.code_check(r):
            self.device_status = 'on' if r.get('result', {}).get(
                'powerSwitch_1') == 1 else 'off'
        else:
            logger.debug('Error getting %s details', self.device_name)

    def turn_on(self) -> bool:
        """Turn BSDGO1 outlet on."""
        return self._set_power(True)

    def turn_off(self) -> bool:
        """Turn BSDGO1 outlet off."""
        return self._set_power(False)

    def _set_power(self, power: bool) -> bool:
        """Set power state of BSDGO1 outlet."""
        body = Helpers.req_body(self.manager, 'bypassV2')
        body['cid'] = self.cid
        body['configModule'] = self.config_module
        body['payload'] = {
            'data': {'powerSwitch_1': 1 if power else 0},
            'method': 'setProperty',
            'source': 'APP'
        }

        r, _ = Helpers.call_api(
            '/cloud/v2/deviceManaged/bypassV2',
            'post',
            headers=Helpers.req_header_bypass(),
            json_object=body,
        )

        if Helpers.code_check(r):
            self.device_status = 'on' if power else 'off'
            return True
        logger.warning('Error turning %s %s', self.device_name, 'on' if power else 'off')
        return False
